/*
 * This file is part of City Domination, which is a GameScript for OpenTTD
 * Copyright (C) 2011  Daniel Beal
 *
 * City Domination is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * City Domination is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with City Domination; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

import("util.superlib","SuperLib",19);
import("util.text", "Text", 1);
Helper <- SuperLib.Helper;
Tile <- SuperLib.Tile;
Log <- SuperLib.Log;

class Town
{
	id = null;
	owner = null;
	constructor(id, owner) {
		this.id = id;
		this.owner = owner;
	}
	function ChangeOwner(newOwner);
}

function Town::ChangeOwner(newOwner)
{
	owner = newOwner;
	Log.Info(GSCompany.GetName(newOwner) + " took over " + GSTown.GetName(id), Log.LVL_SUB_DECISIONS);
	GSNews.Create(GSNews.NT_ECONOMY,Text(GSText.STR_TOWN_TAKEN_OVER, id, newOwner),GSCompany.COMPANY_INVALID);
}

class CityDomination extends GSController
{
	newCompanies = null;
	towns = null;
	constructor() {
		newCompanies = GSList();
		towns = [];
	}
	function Start();
	function CheckEvents();
	function CheckMap();
	function UpdateTownOwners();
}

function CityDomination::Start()
{
	local townlist = GSTownList();
	foreach(town, _ in townlist) towns.append(Town(town, GSCompany.COMPANY_INVALID));
	UpdateTownOwners();
	Log.Info("City Domination Started!", Log.LVL_SUB_DECISIONS);
	GSController.Sleep(1);
	while(true) {
		CheckEvents();
		CheckMap();
		GSController.Sleep(10);
	}
}

function CityDomination::CheckEvents()
{
	while(GSEventController.IsEventWaiting()) {
		local event = GSEventController.GetNextEvent();
		if(event == null) continue;
		local eventType = event.GetEventType();
		if(eventType == GSEvent.ET_COMPANY_BANKRUPT) {
			// If the company owned towns, set them to no owner
			Log.Info("Found bankrupt company!", Log.LVL_SUB_DECISIONS);
			local deadcompany = GSEventCompanyBankrupt.Convert(event);
			foreach(town in towns) {
				if(town.owner == deadcompany.GetCompanyID()) town.owner == GSCompany.COMPANY_INVALID;
			}
		}
		if(eventType == GSEvent.ET_COMPANY_NEW) {
			Log.Info("Found new company!", Log.LVL_SUB_DECISIONS);
			local newcompany = GSEventCompanyNew.Convert(event);
			newCompanies.AddItem(newcompany.GetCompanyID(),0);
		}
	}
}

function CityDomination::CheckMap()
{
	foreach(company, _ in newCompanies) {
		if(GSCompany.GetCompanyHQ(company) != GSMap.TILE_INVALID) {
			Log.Info("Company built HQ: " + GSCompany.GetName(company), Log.LVL_SUB_DECISIONS);
			GSSign.BuildSign(GSCompany.GetCompanyHQ(company),Text(GSText.STR_HQ,company));
			GSNews.Create(GSNews.NT_GENERAL, Text(GSText.STR_NEW_HQ, company, GSTile.GetClosestTown(GSCompany.GetCompanyHQ(company))), GSCompany.COMPANY_INVALID);
			newCompanies.RemoveItem(company);
		}
	}
	foreach(town in towns) {
		// Get a list of all stations inside the town
		local tilelist = Tile.GetTownTiles(town.id);
		tilelist.Valuate(GSTile.IsStationTile);
		tilelist.KeepValue(1);
		local stationlist = GSList();
		foreach(tile, _ in tilelist) {
			if(!stationlist.HasItem(GSStation.GetStationID(tile))) {
				Log.Info("Adding station " + GSStation.GetName(GSStation.GetStationID(tile)));
				stationlist.AddItem(GSStation.GetStationID(tile), GSStation.GetCargoRating(GSStation.GetStationID(tile), Helper.GetPAXCargo()));
			}
		}
		if(stationlist.IsEmpty()) {
//			Log.Info("Town " + GSTown.GetName(town) + " has no stations in it");
			continue;
		}
		// Now get a list of companies get a score for each of them
		local companylist = GSList();
		for(local competitor = GSCompany.COMPANY_FIRST;competitor <= GSCompany.COMPANY_LAST;competitor++) {
			if(GSCompany.ResolveCompanyID(competitor) != GSCompany.COMPANY_INVALID) companylist.AddItem(competitor,0);
		}
		foreach(station, score in stationlist) {
			companylist.SetValue(GSStation.GetOwner(station), companylist.GetValue(GSStation.GetOwner(station)) + score);
		}
		// Finally, determine the company that should own the city
		companylist.Sort(GSList.SORT_BY_VALUE, false);
		local oldowner = town.owner;
		Log.Info("Old Owner: " + GSCompany.GetName(oldowner));
		if(oldowner != companylist.Begin()) town.ChangeOwner(companylist.Begin());
		Log.Info("New Owner: " + GSCompany.GetName(companylist.Begin()));
		if(oldowner != town.owner) {
			UpdateTownOwners(town);
		}
		Log.Info("Winning Company: " + GSCompany.GetName(town.owner));
		GSController.Sleep(2);
	}
}

function CityDomination::UpdateTownOwners(town = null)
{
	Log.Info("Updating town orders...", Log.LVL_SUB_DECISIONS);
	if(!town) {
		Log.Info("...without specific index!", Log.LVL_SUB_DECISIONS);
		foreach(atown in towns) {
			local signlist = GSSignList();
			Log.Info("Sign List Length: " + signlist.Count());
			signlist.Valuate(GSSign.GetLocation);
			signlist.KeepValue(GSTown.GetLocation(atown.id));
			Log.Info("Sign List Length: " + signlist.Count());
			foreach(sign in signlist) GSSign.RemoveSign(sign);
			local signid;
			if(GSCompany.ResolveCompanyID(atown.owner) == GSCompany.COMPANY_INVALID) signid = GSSign.BuildSign(GSTown.GetLocation(atown.id), Text(GSText.STR_NOT_OWNED));
			else signid = GSSign.BuildSign(GSTown.GetLocation(atown.id), Text(GSText.STR_OWNER, GSCompany.GetName(atown.owner)));
			// GSSign.BuildSign(GSTown.GetLocation(atown.id), Text(GSText.STR_NOT_OWNED));
			if(!GSSign.IsValidSign(signid)) Log.Warning("Failed to build sign: " + GSError.GetLastErrorString());
			GSController.Sleep(1);
		}
	}
	else {
		local signlist = GSSignList();
		Log.Info("Sign List Length: " + signlist.Count());
		signlist.Valuate(GSSign.GetLocation);
		signlist.KeepValue(GSTown.GetLocation(town.id));
		Log.Info("Sign List Length: " + signlist.Count());
		foreach(sign in signlist) GSSign.RemoveSign(sign);
		if(GSCompany.ResolveCompanyID(town.owner) == GSCompany.COMPANY_INVALID) GSSign.BuildSign(GSTown.GetLocation(Text(GSText.STR_NOT_OWNED)));
		else GSSign.BuildSign(GSTown.GetLocation(town.id), Text(GSText.STR_OWNER, town.owner));
		GSController.Sleep(1);
	}
}
